import{j as r}from"./@tanstack-CEbkxrhX.js";import{s as n,P as i,az as m}from"./index-BawkpTlr.js";const o=m("flex items-center text-text-sm rounded-md border px-4 py-3",{variants:{intent:{primary:"border-primary-400 bg-primary-25",neutral:"border-theme-border-moderate"}},defaultVariants:{intent:"primary"}});function l({children:e,className:t,intent:a,...s}){return r.jsxs("div",{...s,className:n(o({intent:a}),t),children:[r.jsx(i,{className:"mr-4 h-5 w-5 shrink-0 fill-theme-text-brand"}),r.jsx("div",{className:"w-full min-w-0",children:e})]})}export{l as I};
