import{j as e}from"./@tanstack-CEbkxrhX.js";import{$ as o,a0 as a,h as r,a1 as c,a3 as d,a6 as n,a7 as m}from"./index-BawkpTlr.js";import{S as x}from"./play-circle-DK5QMJyp.js";import{C as h}from"./CodeSnippet-WEzpO0az.js";function f({videoLink:s,isButton:t=!0,buttonText:l,fallbackImage:i}){return e.jsxs(o,{children:[t?e.jsx(a,{asChild:!0,children:e.jsxs(r,{className:"mt-5 h-auto gap-1 px-2 py-1 sm:h-7",size:"md",children:[e.jsx(x,{className:"h-5 w-5 shrink-0 fill-white"}),l??e.jsx(e.Fragment,{children:"Watch the Quickstart Guide (3 min)"})]})}):e.jsx(a,{children:i}),e.jsxs(c,{className:"max-w-[1000px]",children:[e.jsx("div",{className:"flex items-center justify-between border-b border-theme-border-moderate py-2 pl-5 pr-3",children:e.jsx(d,{className:"text-text-lg",children:"Get Started"})}),e.jsx("div",{className:"",children:e.jsx("iframe",{className:"aspect-video w-full overflow-hidden",src:s,allowFullScreen:!0,allow:"fullscreen"})}),e.jsx(n,{children:e.jsx(m,{asChild:!0,children:e.jsx(r,{size:"md",children:"Close"})})})]})]})}function w(s){return e.jsxs("div",{className:"space-y-1",children:[e.jsx("p",{className:"text-text-sm text-theme-text-secondary",children:s.description}),e.jsx(h,{codeClasses:"whitespace-pre-wrap",wrap:!0,code:s.command})]})}export{f as V,w as g};
