import{j as e}from"./@tanstack-CEbkxrhX.js";import{P as r}from"./PageHeader-B0pUife2.js";import{B as a,aw as n,p as l}from"./index-BawkpTlr.js";import{C as i}from"./CodeSnippet-WEzpO0az.js";import{I as m}from"./Infobox-C7bf70VS.js";import{b as d}from"./@react-router-DYovave8.js";import{C as c,m as x}from"./cloud-only-Bkawp7CJ.js";import{r as p}from"./@radix-C9DBgJhe.js";import"./@reactflow-DYIyhCfd.js";import"./copy-BRhQz3j-.js";function u(){return e.jsx(m,{children:e.jsxs("div",{className:"flex w-full flex-wrap items-center gap-x-2 gap-y-0.5 text-text-md",children:[e.jsx("p",{className:"font-semibold",children:"This is a ZenML Cloud feature. "}),e.jsx("p",{children:"Upgrade to ZenML Cloud to access the Model Control Plane and interact with your models in the Dashboard."})]})})}function f(){const[o]=d(),s=o.get("model");function t(){return s?`zenml model list --name='contains:${s}'`:"zenml model list"}return e.jsxs(a,{className:"flex flex-wrap justify-between p-5",children:[e.jsxs("div",{children:[e.jsx("p",{className:"text-text-xl font-semibold",children:"Staying Open Source? "}),e.jsx("p",{className:"text-theme-text-secondary",children:"No problem! Use this CLI command to list your models"})]}),e.jsx(i,{code:t()})]})}const h="/assets/mcp-Cb1aMeoq.webp";function I(){const{setTourState:o,tourState:{tourActive:s}}=n();return p.useEffect(()=>{s&&o(t=>({...t,run:!0,stepIndex:t.stepIndex}))},[s]),e.jsxs("div",{children:[e.jsxs(r,{className:"flex items-center gap-1",children:[e.jsx("h1",{className:"text-display-xs font-semibold",children:"Models"}),e.jsx(l,{color:"purple",rounded:!0,size:"sm",children:e.jsx("span",{className:"font-semibold text-primary-500",children:"Cloud"})})]}),e.jsxs("div",{className:"layout-container space-y-5 py-5",children:[e.jsx(u,{}),e.jsx(c,{image:{src:h,alt:"Screenshot of the ZenML Cloud Model Control plane"},features:x}),e.jsx(f,{})]})]})}export{I as default};
