import{j as e}from"./@tanstack-CEbkxrhX.js";import{B as l,ay as r}from"./index-BawkpTlr.js";const s=["Managed ZenML server on your VPC or hosted on our servers","Social SSO, RBAC, and User Management","CI/CD/CT, Artifact Control Plane and more!"],c=["Model Control Plane Dashboard",...s],m=["Artifact Control Plane Dashboard",...s];function x({features:t,image:a}){return e.jsxs(l,{className:"relative overflow-hidden px-7 py-5",children:[e.jsxs("div",{className:"max-w-[450px] space-y-4",children:[e.jsx("p",{className:"text-display-xs font-semibold",children:"Access Advanced Model Management Features with ZenML Cloud"}),e.jsx("ul",{className:"space-y-2",children:t.map((n,o)=>e.jsx("li",{className:"text-text-md text-theme-text-secondary",children:n},o))}),e.jsxs("div",{className:"flex flex-wrap items-center gap-2",children:[e.jsx("a",{href:"https://zenml.io/pricing",target:"_blank",rel:"noopener noreferrer",className:r({intent:"primary",size:"md"}),children:"Start the free trial"}),e.jsx("a",{href:"https://www.zenml.io/cloud",target:"_blank",rel:"noopener noreferrer",className:r({intent:"primary",size:"md",emphasis:"minimal"}),children:"Learn more"})]})]}),e.jsx("img",{className:"absolute right-0 top-0 hidden translate-x-[45%] translate-y-[10%] rounded-md border border-theme-border-moderate lg:block",src:a.src,alt:a.alt})]})}export{x as C,m as a,c as m};
